//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <cassert>

namespace Tools
{
	//-------------------------------------------------------------------------
	// Memory Pool Definition
	//-------------------------------------------------------------------------
	template<typename T>
	MemoryPool<T>::MemoryPool(std::size_t _numBlockPerChunk):
	blockSize(sizeof(T)),
	numBlocks(_numBlockPerChunk),
	allocationChunk(NULL), 
	deallocationChunk(NULL)
	{
		
	}
	//-------------------------------------------------------------------------
	template<typename T>
	MemoryPool<T>::~MemoryPool()
	{
		typename TChunkCollection::iterator it = chunks.begin();
		for (; it != chunks.end(); ++it)
		{
		   assert(it->BlocksAvailable == numBlocks); // Non free memory !
		   it->Release();
		}
	}
	//-------------------------------------------------------------------------
	template<typename T>
	void* MemoryPool<T>::Allocate() 
	{ 
		printf("MemoryPool::Allocate\n");
		if (allocationChunk == 0 || allocationChunk->BlocksAvailable == 0) 
		{
			// No available memory in this chunk 
			// Try to find one 
			typename TChunkCollection::iterator i = chunks.begin(); 
			for (;; ++i) 
			{
				if (i == chunks.end()) 
				{ 
					// All filled up-add a new chunk 
					chunks.reserve(chunks.size()+1); 
					Chunk newChunk; 
					newChunk.Initialize(blockSize, numBlocks); 
					chunks.push_back(newChunk); 
					allocationChunk = &chunks.back(); 
					deallocationChunk = &chunks.back(); 
					break; 
				} 

				if (i->BlocksAvailable > 0) 
				{ 
					// Found a chunk 
					allocationChunk = &*i; 
					break; 
				} 
			} 
		} 

		assert(allocationChunk != 0); 
		assert(allocationChunk->BlocksAvailable > 0); 
		return allocationChunk->Allocate(blockSize);
	}
	//-------------------------------------------------------------------------
	template<typename T>
	void MemoryPool<T>::Desallocate(void* _ptr)
	{ 
		printf("MemoryPool::Deallocate\n");
	 	assert(!chunks.empty());
		assert(&chunks.front() <= deallocationChunk);
		assert(&chunks.back() >= deallocationChunk);
		
		deallocationChunk = Find(_ptr);
		assert(deallocationChunk);

		DoDeallocate(_ptr);
	}
	//-------------------------------------------------------------------------
	template<typename T>
	void MemoryPool<T>::DoDeallocate(void* _ptr)
	{
		assert(deallocationChunk->Data <= _ptr);
		assert(deallocationChunk->Data + numBlocks * blockSize > _ptr);

		// call into the chunk, will adjust the inner list but won't release memory
		deallocationChunk->Deallocate(_ptr, blockSize);

		if (deallocationChunk->BlocksAvailable == numBlocks)
		{
			// deallocChunk_ is completely free, should we release it? 
			Chunk& lastChunk = chunks.back();

			if (&lastChunk == deallocationChunk)
			{
				// check if we have two last chunks empty
				if (chunks.size() > 1 && 
				deallocationChunk[-1].BlocksAvailable == numBlocks)
				{
					// Two free chunks, discard the last one
					lastChunk.Release();
					chunks.pop_back();
					allocationChunk = deallocationChunk = &chunks.front();
				}
				return;
			}

			if (lastChunk.BlocksAvailable == numBlocks)
			{
				// Two free blocks, discard one
				lastChunk.Release();
				chunks.pop_back();
				allocationChunk = deallocationChunk;
			}
			else
			{
				// move the empty chunk to the end
				std::swap(*deallocationChunk, lastChunk);
				allocationChunk = &chunks.back();
			}
		}
	}
	//-------------------------------------------------------------------------
	template<typename T>
	typename MemoryPool<T>::Chunk* MemoryPool<T>::Find(void* _ptr)
	{
		assert(!chunks.empty());
		assert(deallocationChunk);

		const std::size_t chunkLength = numBlocks * blockSize;

		Chunk* lo = deallocationChunk;
		Chunk* hi = deallocationChunk + 1;
		Chunk* loBound = &chunks.front();
		Chunk* hiBound = &chunks.back() + 1;

		// Special case: deallocationChunk is the last in the array
		if (hi == hiBound) hi = 0;

		for (;;)
		{
			if (lo)
			{
				if (_ptr >= lo->Data && _ptr < lo->Data + chunkLength)
				{
					return lo;
				}
				if (lo == loBound) 
					lo = 0;
				else 
					--lo;
			}

			if (hi)
			{
				if (_ptr >= hi->Data && _ptr < hi->Data + chunkLength)
				{
					return hi;
				}
				if (++hi == hiBound) 
					hi = 0;
			}
		}
		assert(false);
		return 0;
	}

	//-------------------------------------------------------------------------
	// Chunk definition
	//-------------------------------------------------------------------------
	template<typename T>
	void MemoryPool<T>::Chunk::Initialize(std::size_t _blockSize, Byte _blocks)
	{ 
		assert(_blockSize > 0);
		assert(_blocks > 0);
		assert((_blockSize * _blocks) / _blockSize == _blocks); // Overflow check

		Data = new Byte[_blockSize * _blocks]; 
		FirstAvailableBlock = 0; 
		BlocksAvailable = _blocks; 
		unsigned char i = 0; 
		unsigned char* ptr = Data; 
		for (; i != _blocks; ptr += _blockSize) 
		{ 
			*ptr = ++i; 
		} 
	} 
	//-------------------------------------------------------------------------
	template<typename T>
	void MemoryPool<T>::Chunk::Release()
	{ 
		delete[] Data;
	} 
	//-------------------------------------------------------------------------
	template<typename T>
	void* MemoryPool<T>::Chunk::Allocate(std::size_t _blockSize) 
	{ 
		if (!BlocksAvailable) return 0; 

	    assert((FirstAvailableBlock * _blockSize) / _blockSize == FirstAvailableBlock);

		Byte* ptr = Data + (FirstAvailableBlock * _blockSize); 

		// Update firstAvailableBlock_ to point to the next block 
		FirstAvailableBlock = *ptr; 
		--BlocksAvailable;
		return ptr; 
	}
	//-------------------------------------------------------------------------
	template<typename T>
	void MemoryPool<T>::Chunk::Deallocate(void* _ptr, std::size_t _blockSize) 
	{ 
		assert(_ptr >= Data); 
		Byte* toRelease = static_cast<Byte*>(_ptr); 

		// Alignment check 
		assert((toRelease - Data) % _blockSize == 0); 
		*toRelease = FirstAvailableBlock; 
		FirstAvailableBlock = static_cast<Byte>((toRelease - Data) / _blockSize); 

		// Truncation check 
		assert(FirstAvailableBlock == (toRelease - Data) / _blockSize); 
		++BlocksAvailable; 
	} 
}
